/***********************************************************/
/* WCPsimul: a Web Cache removal Policy simulation tool    */
/* Copyright (C) 1997 Nicolas Niclausse                    */
/*                                                         */
/* This program is free software; you can redistribute     */
/* it and/or modify it under the terms of the GNU General  */
/* Public License as published by the Free Software        */
/* Foundation; either version 2 of the License, or (at     */
/* your option) any later version.                         */
/*                                                         */
/* This program is distributed in the hope that it will    */
/* be useful, but WITHOUT ANY WARRANTY; without even the   */
/* implied warranty of MERCHANTABILITY or FITNESS FOR A    */
/* PARTICULAR PURPOSE.  See the GNU General Public License */
/* for more details.                                       */
/*                                                         */
/* You should have received a copy of the GNU General      */
/* Public License along with this program; if not, write   */
/* to the Free Software Foundation, Inc., 675 Mass Ave,    */
/* Cambridge, MA 02139, USA.                               */
/*                                                         */
/* For information, you can contact me at:                 */
/* Nicolas.Niclausse@sophia.inria.fr                       */
/***********************************************************/

// Classe Cache

// Le cache est compose d'une liste chainee (classe Cache) contenant
// les references des documents. On assigne un cout a chaque
// document. Le garbage collector est active lorsque le taux de
// remplissage du cache depasse hit_high (90% par defaut). Les
// documents sont effacees par ordre croissant de leur fonction de
// cout (on garde en priorite les document les plus couteux), jusqu'a
// ce que le taux de remplissage soit devenu inferieur a hit_low (75%
// par defaut). Pour realiser le tri par ordre croissant du cout des
// documents, on utilise un tableau (Cache.index_list) qui contient
// des pointeurs sur les documents (classe Log_URL).  La politique est
// donc entierement determinee par la methode cost de la classe
// Log_URL Utilisation d'un table de hachage (class Hash_table) afin
// d'optimser la boucle de recherche d'URL. (Ameliore *beaucoup* les
// performances)

#include <iostream.h>
#include <DLList.h>
#include <Pix.h>
#include "TTableau.h"
#include "Hash.h"
#include <String.h>
#include <fstream.h>
#include <std.h>
#include <GetOpt.h>
#include "WCPsimul.h"
#include <sys/time.h>
#include <math.h>
#include <limits.h>

#include "Cache.h"

#define min(a,b) (a>b? b : a)
#define CONN_INIT 1024

void Cache::init()
{
  mem_libre=0;
  tot_mem_libre=0;
  obj_max_size=0; 
  hit_low=0.90;
  hit_high=0.95;
  miss=0;
  hit=0;
  newsize=0;
  w_miss=0;
  w_hit=0;
  ws_miss=0;
  gccount = 0;
  n_overflow = 0;
  total_ref = 0;
  total_time = 0;
  debug = 0;
  high_low_flag = 0;
  politique = 'r';
  nombre_objets = 0;
  start_recording = 0;
  nombre_max_objets = URL_MAX;
  debit_hit = 100;
  index_list.resize(nombre_max_objets); 
};

Cache::Cache()
{
  init();
};

Cache::Cache(double mem,double object)
{
  init();
  mem_libre=mem;
  tot_mem_libre=mem;
  obj_max_size=object; 
};
  
Cache::Cache(double mem, double low, double high, double object, int h_l_flag, double deb_hit, int start)
{
  init();
  mem_libre=mem;
  tot_mem_libre=mem;
  obj_max_size=object; 
  hit_low=low;
  hit_high=high;
  high_low_flag=h_l_flag;
  debit_hit=deb_hit;
  start_recording = start;
  // 1 -> on commence tout de suite
  // 0 -> on commence apres premier garbage collect
};

void Cache::output(ofstream& out)
{
  long i;

  out.precision(15);
  for (i=0;i<nombre_objets;i++)
    {
      out << index_list[i]->date_last_ref << " ";
      out << index_list[i]->retrieve_time << " ";
      out << "SIMU ";
      out << index_list[i]->frequency << " ";
      out << index_list[i]->size << " ";
      out << "# ";
      out << index_list[i]->URL << "\n";
    }
};

int Cache::active_gc()
{
  if (high_low_flag)
    return 0;
  else
    return ((1-mem_libre/tot_mem_libre) > hit_high);
};

int compare_function(Log_URL*& i1, Log_URL*& i2)
{
  if (i1->cost < i2->cost)
    return 1;
  else  if (i1->cost == i2->cost)
    {
      if (i1->cost2 < i2->cost2)
	return 1;
      else if (i1->cost2 == i2->cost2)
	return 0;
      else
	return -1;
    }
  else
    return -1;
}

void Cache::insert_hit(Log_URL& url, Log_URL *& indice)
{
  if (debug) cerr << "# hit "<< url.size <<" " << url.URL <<"\n";
  indice->frequency++;
  total_ref++;
  total_time += url.retrieve_time; // si start-recording ???
  indice->date_last_ref = url.date_last_ref;
  indice->retrieve_time = url.retrieve_time;
  //  doit on modifier le retrieve_time ?
  if (start_recording==1)
    {
      hit++;
      w_hit+=url.size/debit_hit;
    }
  if (debug) cerr << "#time-miss: 0\n";
}

double Cache::cost(Log_URL *& i)
{
  return i->date_last_ref;
}

double Cache::cost2(Log_URL *& i)
{
  return double(i->frequency);
}

void Cache::insert_newsize(Log_URL& url, Log_URL *& indice)
{
  double diff = 0;
  
  if (debug) cerr << "# miss(newsize) "<< url.size <<" " << url.URL <<"\n";
  diff = indice->size-url.size;
  mem_libre+=diff;
  if ((mem_libre < 0) && (debug))
    {
      cerr << "Warning !:Memoire libre nagative:"<< mem_libre <<"\n";
      url.print();
      indice->print();
    }
  if (mem_libre<0)
    n_overflow++;
  indice->frequency++;
  total_ref++;
  total_time += url.retrieve_time;
  indice->size = url.size;
  indice->retrieve_time = url.retrieve_time;
  indice->date_last_ref = url.date_last_ref;
  //   cache_list(indice).date_first_ref=url.date_last_ref;


  if (start_recording==1)
    {
      miss++;
      w_miss += url.retrieve_time;
      ws_miss += url.size;
  //  insert_hit(url,indice);
    }
  newsize++;
  if (debug) cerr << "#time-miss: "<< url.retrieve_time << "\n";
}

void Cache::overflow_miss(Log_URL& url)
{
  if (debug) cerr << "# overflow_miss "<< url.size <<" " << url.URL <<"\n";
  if (start_recording==1)
    {
      miss++;
      w_miss += url.retrieve_time;
      ws_miss += url.size;
    }
  total_ref++;
  total_time += url.retrieve_time;
}

void Cache::insert_miss(Log_URL& url)
{
  Log_URL * tmp;

  tmp = &url;
  

  if (debug) cerr << "# miss "<< url.size <<" " << url.URL <<"\n";
  
  url.frequency = 1;
  total_ref++;
  total_time += url.retrieve_time;
  url.persistent = 'n';
  url.date_first_ref = url.date_last_ref;
  index_list[nombre_objets] = hash.insert(url);
  nombre_objets++;
  mem_libre -= url.size;
  
  if (start_recording == 1)
    {
      miss++;
      w_miss += url.retrieve_time;
      ws_miss += url.size;
    }
  //  if (mem_libre<0)
  //    n_overflow++;
  if (debug) cerr << "#time-miss: "<< url.retrieve_time << "\n";
}

void Cache::insert_init(Log_URL& url)
{
  url.frequency = 1;
  total_ref++;
  url.date_first_ref = url.date_last_ref;
  index_list[nombre_objets] = hash.insert(url);
  nombre_objets++;
  mem_libre -= url.size;
  start_recording = 1; // si fichier d'initialisation, on commence les mesures tout de suite

  if (mem_libre < 0)
    n_overflow++;
}

void Cache::update_cost()
{
  // TTableau<Log_URL*> index_list[];
  for (long i = 0; i<nombre_objets; i++)
    {
      index_list[i]->cost = cost(index_list[i]);
      index_list[i]->cost2 = cost2(index_list[i]);
    }
}

void Cache::gc(double actual_time, const Log_URL& url)
// Garbage Collector 
{
  double max = 0;
  long i = nombre_objets-1;
  Log_URL * tmp;

  time = actual_time;

  // Si premier garbage collect -> on commence a enregistrer les stats 
  if (start_recording == 0)
    start_recording = 1;

  
  if (high_low_flag)
    max = url.size;
  else
    max = (1-hit_low)*tot_mem_libre;
  
  if (debug) cerr << "# Garbage Collect: " << mem_libre <<"\n";
  gccount++;

     {
       update_cost();
       index_list.sort(compare_function,nombre_objets);

        if (debug)
  	 for(long j=0;j<nombre_objets;j++)
 	   {
  	     cerr << index_list[j]->URL<<" "<< index_list[j]->cost << " ";
 	   }
        if (debug)  cout << "\n ";
       
       // tant que la memoire libre n'est pas inferieure au seuil max (mecanisme utilise dans Squid) 
       
       tmp = index_list[i];
       while (mem_libre < max)
	 {
	   mem_libre += tmp->size;
	   total_ref -= tmp->frequency;
	   hash.remove(tmp->URL);
	   i--;
	   tmp = index_list[i];
	 }
       if (i != (nombre_objets-1))
	 nombre_objets = i+1;
     }
   if (debug) cerr << "# Fin Garbage Collect: " << mem_libre <<"\n";
}

// ************************************************************************
//
//                             Cache_aw
//
// ************************************************************************

Cache_aw::Cache_aw() : serveur(couple_init)
{
  CONN = CONN_INIT;
  ALPHA = 0.125;
  W_B = 8196;
  W_N = 0.9;
};

Cache_aw::Cache_aw(double mem, double low, double high, double object, int h_l_flag, double deb_hit, long conn, long w_b, double w_n, int start): Cache( mem,  low,  high,  object,  h_l_flag, deb_hit, start),serveur(couple_init)
{
  CONN = conn;
  ALPHA = 0.125;
  W_B = w_b;
  W_N = w_n;
};

void Cache_aw::update_server_stat(Couple & old, const Couple & sample)
{
  // sample contient echantillon de latence (si size < CONN) ou debit (sinon) 
  
  if (sample.debit > 0) // size > CONN
    old.debit += ALPHA*(sample.debit-old.debit);
  else
    old.latence += ALPHA*(sample.latence-old.latence);
};

Couple Cache_aw::init_couple(double size, double time, const String & serveur_name)
{
  Couple update;

  if (size<CONN)
    {
      update.debit = -1;
      update.latence = time;
    }
  else
    {
      if (serveur.contains(serveur_name))
      	update.debit = (size-CONN) / (time-serveur[serveur_name].latence);
      else
	update.debit = size/time;
      update.latence = -1;
    }
  return update;
};

double Cache_aw::cost(Log_URL *& i)
{
  String nom = get_server_name(i->URL);
  Couple sample = serveur[nom];
    
   if (sample.debit <= 0) // si on n'as pas d'estimation du debit pour ce serveur
     //     return double(i->retrieve_time); // on renvoit le temps mis precedement
     return (W_B/double(i->retrieve_time))*(pow(double(i->frequency), W_N));
   else
     //     return sample.latence+double(i->size)/sample.debit;
     return (sample.latence+W_B/sample.debit)*(pow(double(i->frequency), W_N))/double(i->size);
}

void Cache_aw::insert_newsize(Log_URL& url, Log_URL *& indice)
{
   String nom = get_server_name(url.URL);
   Couple sample = init_couple(url.size, url.retrieve_time, nom);

  Cache::insert_newsize(url,indice);
  update_server_stat(serveur[nom], sample);
}


void Cache_aw::overflow_miss(Log_URL& url)
{
  String nom = get_server_name(url.URL);
  Couple sample = init_couple(url.size, url.retrieve_time, nom);

  Cache::overflow_miss(url);
  update_server_stat(serveur[nom], sample);
}

void Cache_aw::insert_miss(Log_URL& url)
{
  String nom = get_server_name(url.URL);
  Couple sample = init_couple(url.size, url.retrieve_time, nom);
  
  Cache::insert_miss(url);
  update_server_stat(serveur[nom], sample);
}

void Cache_aw::insert_hit(Log_URL & url, Log_URL *& indice)
{
  Cache::insert_hit(url, indice);
  
  if (debug) cerr << "nombre de serveurs: "<<serveur.length() <<" "<< sizeof(serveur)<<"\n";
}

void Cache_aw::gc(double time, const Log_URL& url)
{
  Cache::gc(time, url);
  if (debug) output_serveur();
}

String Cache_aw::get_server_name(const String & URL) const
{
  String * words;
  String result;
  const String delimiteur="/";
  const indice_serveur = 2;
  
  words = new String[20];
  //  int nn = split(URL, words, 10, delimiteur);
  result = words[indice_serveur];
  delete[] words;
  return result;
};

void Cache_aw::output_serveur()
{
  long i;
  Pix ind = serveur.first();
  String k;
  Couple x;
  
  for (i=0;i<serveur.length();i++)
    {
      k = serveur.key(ind);
      x = serveur.contents(ind);
      cerr << k << " " << x.debit<< " " << x.latence << "\n";
      serveur.next(ind);
    }
};

// ************************************************************************
//
// 			Classe Cache_freq
//
// ************************************************************************

double Cache_freq::cost(Log_URL *& i)
{
  return double(i->frequency);
}

double Cache_freq::cost2(Log_URL *& i)
{
  return i->date_last_ref;
}

Cache_freq::Cache_freq(double mem,double object) : Cache(mem, object) {};

Cache_freq::Cache_freq(double mem, double low, double high, double object, int h_l_flag,double deb_hit, int start): Cache( mem,  low,  high,  object,  h_l_flag, deb_hit, start){};

// ************************************************************************
//
// 			Classe Cache_size
//
// ************************************************************************

double Cache_size::cost(Log_URL *& i)
{
  return 1.0 / double(i->size);
}

double Cache_size::cost2(Log_URL *& i)
{
  return i->date_last_ref;
}

Cache_size::Cache_size(double mem,double object) : Cache(mem, object) {};

Cache_size::Cache_size(double mem, double low, double high, double object, int h_l_flag,double deb_hit, int start): Cache( mem,  low,  high,  object,  h_l_flag, deb_hit, start){};

// ************************************************************************
//
// 			Classe Cache_mix
//
// ************************************************************************


double Cache_mix::cost(Log_URL *& i)
{
  if (i->frequency > 1)
    if (((time - i->date_first_ref) * i->size) != 0)
      {
// 	if (debug){
// 	  cerr  << "#ARGH= " <<pow(double(time - i->date_first_ref),ALPHA) << "\n";
// 	  i->print();
// 	  cerr << time<<"\n";
// 	}
	
	//	return (double(i->frequency) * log(i->retrieve_time)/(pow(double(time - i->date_last_ref),ALPHA) * pow(i->size,GAMMA)));
	return (double(i->frequency) * pow(i->retrieve_time,BETA)/(pow(double(time - i->date_last_ref),ALPHA) * pow(i->size,GAMMA)));
      }
    else
	{
	  cerr << "erreur division par 0 \n";
	  exit(1);
	}
  else if (i->frequency == 1)
    if ((i->size > 0) && (time > i->date_first_ref))
      {
// 	if (debug) {
// 	  cerr  << "#ARGH2= " <<pow(double(time - i->date_first_ref),ALPHA) << "\n";
// 	  i->print();
// 	  cerr << time<<"\n";
// 	}
	//	return log(i->retrieve_time)/(pow(double(time - i->date_last_ref),ALPHA)*pow(i->size,GAMMA));
	return pow(i->retrieve_time,BETA)/(pow(double(time - i->date_last_ref),ALPHA)*pow(i->size,GAMMA));
      }
    else
      {
	cerr << "erreur size <=0 ou time <= i->date_first_ref \n";
	exit(1);
      }
  else
    {
      cerr << "erreur frequency <1 \n";
      exit(1);
    }
}

double Cache_mix::cost2(Log_URL *& i)
{
  return i->date_last_ref;
}

Cache_mix::Cache_mix(double mem,double object) : Cache(mem, object)
{
  BETA = 1.0;
  ALPHA = 1.0;
  GAMMA = 1.0;
};

Cache_mix::Cache_mix(double mem, double low, double high, double object, int h_l_flag,double deb_hit, int start, double beta, double alpha, double gamma): Cache( mem,  low,  high,  object,  h_l_flag, deb_hit, start)
{
  BETA = beta;
  ALPHA = alpha;
  GAMMA = gamma;
};


// ************************************************************************
//
//                             Cache_greedy
//
// ************************************************************************

Cache_greedy::Cache_greedy() : Cache() {
  L = 0;
};
Cache_greedy::Cache_greedy(double mem,double object) : Cache(mem, object) {
  L = 0;
};

Cache_greedy::Cache_greedy(double mem, double low, double high, double object, int h_l_flag,double deb_hit, int start): Cache( mem,  low,  high,  object,  h_l_flag, deb_hit, start){
  L = 0;
};

double Cache_greedy::cost_obj(Log_URL & i)
{
  return (1/double(i.size)); //Greddy(size)
}

double Cache_greedy::cost(Log_URL *& i)
{
  return (1/double(i->size)); //Greddy(size)
}

void Cache_greedy::insert_newsize(Log_URL& url, Log_URL *& indice)
{
  Cache::insert_newsize(url,indice);
  indice->cost = L+cost(indice);
}

// overflow ??

void Cache_greedy::insert_miss(Log_URL& url)
{
  url.cost = L + cost_obj(url);
  Cache::insert_miss(url);
}

void Cache_greedy::insert_hit(Log_URL & url, Log_URL *& indice)
{
  Cache::insert_hit(url, indice);
  indice->cost = L + cost(indice);
}

void Cache_greedy::gc(double actual_time, const Log_URL& url)
// Garbage Collector 
{
  double max = 0;
  long i = nombre_objets-1;
  Log_URL * tmp;

  time = actual_time;

  // Si premier garbage collect -> on commence a enregistrer les stats 
  if (start_recording == 0)
    start_recording = 1;

  
  if (high_low_flag)
    max = url.size;
  else
    max = (1-hit_low)*tot_mem_libre;
  
  if (debug) cerr << "# Garbage Collect: " << mem_libre <<"\n";
  gccount++;

     {
       index_list.sort(compare_function,nombre_objets);

        if (debug)
  	 for(long j = 0; j < nombre_objets; j++)
 	   {
  	     cerr << index_list[j]->URL<<" "<< index_list[j]->cost << " ";
 	   }
        if (debug)  cout << "\n ";
       
       // tant que la memoire libre n'est pas inferieure au seuil max (mecanisme utilise dans Squid) 
       
       tmp = index_list[i];
       while (mem_libre < max)
	 {
	   mem_libre += tmp->size;
	   total_ref -= tmp->frequency;
	   L = tmp->cost;
	   hash.remove(tmp->URL);
	   i--;
	   tmp = index_list[i];
	 }
       if (i != (nombre_objets-1))
	 nombre_objets = i+1;
     }
     if (debug) cerr << "# Fin Garbage Collect: " << mem_libre <<" L= "<< L <<"\n";
}
// ************************************************************************
//
//                             Cache_greedy_mix
//
// ************************************************************************

Cache_greedy_mix::Cache_greedy_mix() : Cache_greedy() {
};
Cache_greedy_mix::Cache_greedy_mix(double mem,double object) : Cache_greedy(mem, object) {
};

Cache_greedy_mix::Cache_greedy_mix(double mem, double low, double high, double object, int h_l_flag,double deb_hit, int start): Cache_greedy( mem,  low,  high,  object,  h_l_flag, deb_hit, start){
};

double Cache_greedy_mix::cost_obj(Log_URL & i)
{
  return (double(i.frequency) * pow(i.retrieve_time,0.1)/double(i.size));
}

double Cache_greedy_mix::cost(Log_URL *& i)
{
  return (double(i->frequency) * pow(i->retrieve_time,0.1)/double(i->size));
}





// ************************************************************************
//
//                             Cache_dyn
//
// ************************************************************************

Cache_dyn::Cache_dyn(double mem,double object) : Cache(mem, object) {};

Cache_dyn::Cache_dyn(double mem, double low, double high, double object, int h_l_flag,double deb_hit, int start): Cache( mem,  low,  high,  object,  h_l_flag, deb_hit, start){};

void Cache_dyn::init()
{
  mem_libre=0;
  tot_mem_libre=0;
  obj_max_size=0; 
  hit_low=0.90;
  hit_high=0.95;
  miss=0;
  hit=0;
  newsize=0;
  w_miss=0;
  w_hit=0;
  ws_miss=0;
  gccount = 0;
  n_overflow = 0;
  total_ref = 0;
  total_time = 0;
  debug = 0;
  high_low_flag = 0;
  politique = 'r';
  nombre_objets = 0;
  start_recording = 0;
  nombre_max_objets = URL_MAX;
  debit_hit = 100;
};

double Cache_dyn::cost_obj(Log_URL & i)
{
  return (1/double(i.size)); //Greddy(size)
}

double Cache_dyn::cost(Log_URL *& i)
{
  return (1/double(i->size)); //Greddy(size)
}

Pix Cache_dyn::cherche( Log_URL *& indice )
{
  Log_URL * tmp;
  Pix i = index_list_dyn.first();

  //  if (debug) cerr << "debut cherche\n";
  tmp = index_list_dyn(i);

  if (i == 0)
    {
      cerr << "erreur: appel de seek dans une liste chainee vide";
      exit(1);
    }
  else
    {
      while ((tmp != indice) && (i != index_list_dyn.last()))
	{
	  index_list_dyn.next(i);
	  tmp = index_list_dyn(i);
	}
      if (tmp != indice) //fin de liste
	{
	  return 0;
	}
      else
	{
	  return i;
	}
    }
};

void Cache_dyn::insert_objet( Log_URL *& indice, Pix i)
{
   Log_URL * tmp;
   Pix j;
   
   //   if (debug) cerr << "debut insert_objet\n";
   if (i != 0)
     {
       do
	 {
	   j = i;
	   tmp = index_list_dyn(i);
	   index_list_dyn.next(i);
	 }
       while ((indice->cost < tmp->cost) && (i != 0));
       if (i == 0) //fin de liste
	 {
	   if (indice->cost < tmp->cost)
	     index_list_dyn.append(indice);
	   else
	     index_list_dyn.ins_before(j, indice);
	 }
       else
	 {
	   //	   if (debug) cerr << "i!=0\n";
	  index_list_dyn.ins_before(j, indice);
	 }
     }
   else // Liste vide
     {
       index_list_dyn.prepend(indice);
     }
   //   if (debug) cerr << "fin insert_objet\n";
}

void Cache_dyn::insert_newsize(Log_URL& url, Log_URL *& indice)
{
  Pix i = cherche(indice);

  if (debug) cerr << "newsize cost:" << index_list_dyn(i)->cost <<"\n";
  if (i == 0)
    {
      cerr << "erreur: cherche ne trouve rien";
      exit(1);
    }
  Cache::insert_newsize(url, indice);
  indice->cost = cost(indice);

  index_list_dyn.del(i,1);
  i = index_list_dyn.first();

  insert_objet(indice, i);
};

//overflow ??

void Cache_dyn::insert_hit(Log_URL & url, Log_URL *& indice)
{
  Pix i = cherche(indice);
  
  if (i == 0)
    {
      cerr << "erreur: cherche ne trouve rien";
      exit(1);
    }
  Cache::insert_hit(url, indice);
  indice->cost = cost(indice);

  index_list_dyn.del(i,1);
  i = index_list_dyn.first();

  insert_objet(indice, i);
}

void Cache_dyn::insert_miss(Log_URL& url)
{
  Log_URL * tmp;
  
  Pix i = index_list_dyn.first();
 
  if (debug) cerr << "# miss "<< url.size <<" " << url.URL <<"\n";
  
  url.cost = cost_obj(url);
  url.frequency = 1;
  total_ref++;
  total_time += url.retrieve_time;
  url.persistent = 'n';
  url.date_first_ref = url.date_last_ref;

  tmp = hash.insert(url);
  insert_objet(tmp, i);

  nombre_objets++;
  mem_libre -= url.size;
  
  if (start_recording == 1)
    {
      miss++;
      w_miss += url.retrieve_time;
      ws_miss += url.size;
    }
  if (debug) cerr << "#time-miss: "<< url.retrieve_time << "\n";
}

void Cache_dyn::gc(double actual_time, const Log_URL& url)
  //Garbage Collector 
{
  double max = 0;
  Log_URL * tmp;

  time = actual_time;

  // Si premier garbage collect -> on commence a enregistrer les stats 
  if (start_recording == 0)
    start_recording = 1;

  
  if (high_low_flag)
    max = url.size;
  else
    max = (1-hit_low)*tot_mem_libre;
  
  if (debug) cerr << "# Garbage Collect: " << mem_libre <<"\n";
  gccount++;
  
  {
    if (debug)
      for ( Pix j = index_list_dyn.first(); j != 0; index_list_dyn.next(j))
	{
	  cerr << index_list_dyn(j)->URL<<" "<< index_list_dyn(j)->cost << " ";
	}
    if (debug)  cout << "\n ";
    // tant que la memoire libre n'est pas inferieure au seuil max (mecanisme utilise dans Squid) 

    while (mem_libre < max)
	 {
	   tmp = index_list_dyn.remove_rear();
	   mem_libre += tmp->size;
	   total_ref -= tmp->frequency;
	   if (debug) cerr << tmp->URL <<" " << tmp->cost <<"\n";
	   hash.remove(tmp->URL);
	 }
     }
     if (debug) cerr << "# Fin Garbage Collect: " << mem_libre <<"\n";
}

void Cache_dyn::output(ofstream& out)
{
    Pix i;
  
   out.precision(15);
  for ( i = index_list_dyn.first(); i != 0; index_list_dyn.next(i))
    {
      out << index_list_dyn(i)->date_last_ref << " ";
      out << index_list_dyn(i)->retrieve_time << " ";
      out << "SIMU ";
      out << index_list_dyn(i)->frequency << " ";
      out << index_list_dyn(i)->size << " ";
      out << "# ";
      out << index_list_dyn(i)->URL << "\n";
    }
};
