/***********************************************************/
/* WCPsimul: a Web Cache removal Policy simulation tool    */
/* Copyright (C) 1997 Nicolas Niclausse                    */
/*                                                         */
/* This program is free software; you can redistribute     */
/* it and/or modify it under the terms of the GNU General  */
/* Public License as published by the Free Software        */
/* Foundation; either version 2 of the License, or (at     */
/* your option) any later version.                         */
/*                                                         */
/* This program is distributed in the hope that it will    */
/* be useful, but WITHOUT ANY WARRANTY; without even the   */
/* implied warranty of MERCHANTABILITY or FITNESS FOR A    */
/* PARTICULAR PURPOSE.  See the GNU General Public License */
/* for more details.                                       */
/*                                                         */
/* You should have received a copy of the GNU General      */
/* Public License along with this program; if not, write   */
/* to the Free Software Foundation, Inc., 675 Mass Ave,    */
/* Cambridge, MA 02139, USA.                               */
/*                                                         */
/* For information, you can contact me at:                 */
/* Nicolas.Niclausse@sophia.inria.fr                       */
/***********************************************************/

// Classe Cache

// Le cache est compose d'une liste chainee (classe Cache) contenant
// les references des documents. On assigne un cout a chaque
// document. Le garbage collector est active lorsque le taux de
// remplissage du cache depasse hit_high (90% par defaut). Les
// documents sont effacees par ordre croissant de leur fonction de
// cout (on garde en priorite les document les plus couteux), jusqu'a
// ce que le taux de remplissage soit devenu inferieur a hit_low (75%
// par defaut). Pour realiser le tri par ordre croissant du cout des
// documents, on utilise un tableau (Cache.index_list) qui contient
// des pointeurs sur les documents (classe Log_URL).  La politique est
// donc entierement determinee par la methode cost de la classe
// Log_URL Utilisation d'un table de hachage (class Hash_table) afin
// d'optimser la boucle de recherche d'URL. (Ameliore *beaucoup* les
// performances)

#include <iostream.h>
#include <DLList.h>
#include <Pix.h>
#include "TTableau.h"
#include "Log_URL.h"
#include "Hash.h"
#include <String.h>
#include "WCPsimul.h"
#include "String.Couple.CHMap.h"

#ifndef CACHE_H
#define CACHE_H

#define min(a,b) (a>b? b : a)

// ************************************************************************
//
// 		Classe de Cache Generique avec politique LRU
//
// ************************************************************************

class Cache
{
protected:
  double mem_libre; // taille des objets dans le cache en octet
  double tot_mem_libre; // capacite totale du cache en octet
  double hit_low; // taux de remplissage du cache ou le gc s'arrete
  double hit_high; //taux de remplissage du cache a partir duquel on active le gc
  double obj_max_size; //taille maximum d'un objet en octet
  // pour l'instant, vaut tot_mem_libre/2
  double w_miss;
  double ws_miss;
  double w_hit;
  double total_ref; //somme des frequences d'acces de tout le cache
  double total_time; //somme des dure de toutes les requetes 
  long miss,hit,newsize;
  long n_overflow; // nombre de fois ou le cache est rempli a + de 100%
  TTableau<Log_URL*> index_list; // Beurk: Tableau de taille fixe !
  long nombre_objets; // nombre d'objets dans le cache
  long nombre_max_objets; // nombre maximum d'objets dans le cache
  int debug;
  int high_low_flag;
  char politique; // politique du gc
  Hash_table hash;
  double debit_hit; // debit d'un doc dans le cache
  double time; // date courante (utile dans gc selon politique)
  int start_recording; // on commence a enregistrer les hit une fois que le cache est remplit

  virtual double cost(Log_URL *& i);
  virtual double cost2(Log_URL *& i);
  void update_cost();
  virtual void init();

public:
  Log_URL URL;
  long gccount;
  
  Cache();
  Cache(double mem,double object);
  Cache(double mem, double low, double high, double object, int h_l_flag, double deb_hit, int start);
  virtual  ~Cache(){};         // Destructeur

  
  double HIT_high(){return hit_high;};
  double memoire_libre_tot(){return tot_mem_libre;};
  double memoire_libre(){return mem_libre;};
  double object_max_size(){return obj_max_size;};
  int active_gc();

  virtual void insert_miss(Log_URL& url);
  virtual void insert_init(Log_URL& url);
  virtual void overflow_miss(Log_URL& url);
  virtual void insert_hit(Log_URL& url, Log_URL *& indice);
  virtual void insert_newsize(Log_URL& url, Log_URL *& indice);
  virtual void gc(double actual_time, const Log_URL& url);
  virtual void output(ofstream& out);
  
  int search(Log_URL *& ind){ return hash.search(ind);};
  void debug_on() { debug=1;};
  void debug_off() { debug=0;};
  //  void print();
  //  void update(long n, double time);
  double missrate()  {return double(miss)/double(hit+miss);};
  void meanref()
  {
    cerr <<"# meanref="<< double(total_ref)/double(nombre_objets) << " "<< total_ref<< " "<<nombre_objets <<"\n";
  };
  double newsizerate() { return double(newsize)/double(hit+miss);};
  double w_missrate() { return w_miss/double(hit+miss);};
  double win_time() { return w_miss/total_time;};
  long sampling() { return (hit+miss);};
  double w_hitrate() { return w_hit/double(hit+miss);};
  double ws_missrate() { return ws_miss;};
  long overflow() {return n_overflow;};

};

// ************************************************************************
//
// 		Classe de Cache avec politique LFU
//
// ************************************************************************

class Cache_freq: public Cache
{
protected:
  double cost(Log_URL *& i);
  double cost2(Log_URL *& i);
public:
  virtual  ~Cache_freq(){};         // Destructeur
  Cache_freq(){};
  Cache_freq(double mem,double object);
  Cache_freq(double mem, double low, double high, double object, int h_l_flag, double deb_hit, int start);
};

// ************************************************************************
//
// 		Classe de Cache avec politique SIZE
//
// ************************************************************************

class Cache_size: public Cache
{
protected:
  double cost(Log_URL *& i);
  double cost2(Log_URL *& i);
public:
  virtual  ~Cache_size(){};         // Destructeur
  Cache_size(){};
  Cache_size(double mem,double object);
  Cache_size(double mem, double low, double high, double object, int h_l_flag, double deb_hit, int start );
};

// ************************************************************************
//
// 		Classe de Cache avec politique MIX
//
// ************************************************************************

class Cache_mix: public Cache
{
protected:
  double cost(Log_URL *& i);
  double cost2(Log_URL *& i);
  double BETA ; // poids du temps de latence
  double ALPHA ; // poids de l'age du document
  double GAMMA ; // poids de la taille du document
public:
  Cache_mix(){};
  Cache_mix(double mem,double object);
  Cache_mix(double mem, double low, double high, double object, int h_l_flag, double deb_hit, int start, double beta, double alpha, double gamma);
};


// ************************************************************************
//
// 		Classe de Cache avec politique HYB
//
// ************************************************************************

// Classe descendante: tiens compte du temps d'acces au serveur
// utilise une table de hachage ayant pour cle le nom du serveur et
// comme element un Couple de double latence et debit
// Cf article Wooster & Abrams, Proxy caching that estimates page load delays
// conf WWW6

class Cache_aw: public Cache
{
protected:
  long CONN; // si taille document < CONN -> latence, sinon -> debit
  double ALPHA; // constante de lissage (comme pour estimation RTT dans TCP)
  long W_B ; // poids du debit du serveur
  double W_N ; // poids de la frequence d'acces
  Couple couple_init;
  StringCoupleCHMap serveur; 

  Couple init_couple(double size, double time, const String & serveur_name);
  String get_server_name(const String & URL) const;
  void update_server_stat(Couple & old, const Couple & sample);
  double cost(Log_URL *& i);
  
public:
  virtual  ~Cache_aw(){};         // Destructeur
  Cache_aw();
  Cache_aw(double mem, double low, double high, double object, int h_l_flag, double deb_hit, long conn, long w_b, double w_n, int start);
  
  void gc(double time, const Log_URL & url);
  void output_serveur();
  void insert_miss(Log_URL & url);
  void overflow_miss(Log_URL & url);
  void insert_hit(Log_URL & url, Log_URL *& indice);
  void insert_newsize(Log_URL & url, Log_URL *& indice);
};

// ************************************************************************
//
// 		Classe de Cache avec politique Greedy - Size
//
// ************************************************************************

// Classe descendante:
// Cf article - Cost aware WWW Proxy Caching Algorithms
//              Pei Cao & Sandy Irani

class Cache_greedy: public Cache
{
protected:
  double L ; // surcout

  virtual double cost_obj(Log_URL & i);
  virtual double cost(Log_URL *& i);
  
public:
  virtual  ~Cache_greedy(){};         // Destructeur
  Cache_greedy();
  Cache_greedy(double mem,double object);
  Cache_greedy(double mem, double low, double high, double object, int h_l_flag, double deb_hit, int start );
  
  void gc(double actual_time, const Log_URL & url);
  void insert_miss(Log_URL & url);
  void insert_hit(Log_URL & url, Log_URL *& indice);
  void insert_newsize(Log_URL & url, Log_URL *& indice);
};

// ************************************************************************
//
// 		Classe de Cache avec politique Greedy - Mix
//
// ************************************************************************


class Cache_greedy_mix: public Cache_greedy
{
protected:
  double cost_obj(Log_URL & i);
  double cost(Log_URL *& i);

public:
  virtual  ~Cache_greedy_mix(){};         // Destructeur
  Cache_greedy_mix();
  Cache_greedy_mix(double mem,double object);
  Cache_greedy_mix(double mem, double low, double high, double object, int h_l_flag, double deb_hit, int start );
};

// ************************************************************************
//
// 		Classe de Cache avec URL tries dynamiquement
//
// ************************************************************************

class Cache_dyn: public Cache
{
protected:
  
  virtual double cost_obj(Log_URL & i);
  virtual double cost(Log_URL *& i);
  void init();
  void insert_objet(Log_URL *& indice, Pix i);
  DLList < Log_URL * > index_list_dyn;
  Pix cherche(Log_URL *& tmp);
  
public:
  virtual  ~Cache_dyn(){};         // Destructeur
  Cache_dyn();
  Cache_dyn(double mem,double object);
  Cache_dyn(double mem, double low, double high, double object, int h_l_flag, double deb_hit, int start );
  //  ~Cache_dyn(){};
  
  virtual void output(ofstream& out);
  virtual void gc(double actual_time, const Log_URL & url);
  virtual void insert_miss(Log_URL & url);
  virtual void insert_hit(Log_URL & url, Log_URL *& indice);
  virtual void insert_newsize(Log_URL & url, Log_URL *& indice);
};

#endif
